# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Module for Vibration features filter plugin

This filter collects readings for configured observation interval, then calculates statistics on these readings
and puts these statistics into a new reading.
"""

import time
import copy
import logging
import uuid
import pandas as pd
import numpy as np
from pandas.io.json import json_normalize
import json
import datetime

from foglamp.common import logger
import filter_ingest

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__, level = logging.DEBUG)

# Filter specific objects
the_callback = None
the_ingest_ref = None
interval = None
shutdown_in_progress = False
data_window = []
window_start_ts = None

num_windows = 0
total_time = 0

ftr_cols = ['median', 'mean', 'std_dev', 'variance', 'RMS', 'kurtosis']

# FOGL-3453: Mean, Median, Mode, Standard Deviation, Min, Max, Variance, Range

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Vibration features filter plugin',
        'type': 'string',
        'default': 'vibration_features',
        'readonly': 'true'
    },
    'enable': {
        'description': 'Enable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "3"
    },
    'interval': {
        'description': 'Observation time interval over which to calculate features',
        'type': 'float',
        'default': '10',
        'displayName': 'Observation interval (ms)',
        'order': "2"
    },
    'assetName': {
        'description': 'Asset name for reading with calculated features',
        'type': 'string',
        'default': 'VibrationFeature',
        'displayName': 'Asset name',
        'order': "1"
    }
}


def compute_features(handle, dw):
    """ Compute features

    Args:
        A reading data
    """

    # _LOGGER.debug("dw={}".format(dw))

    t1 = datetime.datetime.now()
    df = pd.DataFrame(json_normalize(dw))
    t2 = datetime.datetime.now()
    delta = t2 - t1
    _LOGGER.info("JSON parsing & dataframe creation took {} msec".format(int(delta.total_seconds() * 1000)))

    _LOGGER.debug("Data window has {} rows, TS range = {} .. {}".format(df.shape[0], df['user_ts'][0], df['user_ts'].iloc[-1]))

    feature_reading = {'asset': handle['assetName'], 'timestamp': df['user_ts'].iloc[-1],
                        'key': str(uuid.uuid4()), 'readings': {}}

    t1 = datetime.datetime.now()
    for col in df.columns:
        if col.startswith('readings.ch1'):
            series = pd.Series(df[col])
            ftrs = get_features(series)
            del series
            key = col.replace("readings.", "")
            for k, v in zip (ftr_cols, ftrs):
                feature_reading['readings'][key+'_'+k] = v

    t2 = datetime.datetime.now()
    delta = t2 - t1
    _LOGGER.info("Features calculated in {} msec".format(int(delta.total_seconds() * 1000)))
    _LOGGER.debug("feature_reading={}".format(json.dumps(feature_reading)))

    del df

    return feature_reading


def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        'name': 'vibration_features',
        'version': '1.8.2',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref:
        callback:
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    data = copy.deepcopy(config)

    global the_callback, the_ingest_ref, interval

    the_callback = callback
    the_ingest_ref = ingest_ref
    data['interval'] = float(config['interval']['value'])
    interval = data['interval']
    data['assetName'] = config['assetName']['value']

    _LOGGER.debug("plugin_init for filter vibration_features called")

    return data


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    global interval

    interval = float(new_config['interval']['value'])
    assetName = new_config['assetName']['value']
    _LOGGER.debug("Old config for vibration_feature plugin {} \n new config {}".format(handle, new_config))
    new_handle = copy.deepcopy(new_config)
    new_handle['assetName'] = assetName

    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, interval, data_window
    shutdown_in_progress = True
    time.sleep(1)
    the_callback = None
    the_ingest_ref = None
    interval = None
    data_window = []

    _LOGGER.info('Filter vibration_feature plugin shutdown.')


def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward

    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, data_window, window_start_ts, interval
    if shutdown_in_progress:
        return

    if handle['enable']['value'] == 'false':
        # Filter not enabled, just pass data onwards
        filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data)
        return

    for elem in data:
        user_ts = elem['user_ts']
        if user_ts[-3] == ':':
            user_ts = user_ts[:-3] + user_ts[-2:]

        user_ts = datetime.datetime.strptime(user_ts, '%Y-%m-%d %H:%M:%S.%f%z').timestamp()

        if window_start_ts == None:
            window_start_ts = user_ts

        if user_ts > window_start_ts + interval/1000.0:
            t1 = datetime.datetime.now()

            feature_reading = compute_features(handle, data_window)

            t2 = datetime.datetime.now()

            data_window.append(feature_reading)

            _LOGGER.info("Ingesting {} readings to south service".format(len(data_window)))

            # Pass data onwards
            filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data_window)

            t3 = datetime.datetime.now()

            del data_window, feature_reading
            data_window = []
            window_start_ts = user_ts

            delta1 = t2 - t1
            delta2 = t3 - t2
            delta = t3 - t1
            _LOGGER.info("Data window processed in {} msec: feature calculation took {} msec, ingest took {} msec"
                         .format(int(delta.total_seconds() * 1000), int(delta1.total_seconds() * 1000), int(delta2.total_seconds() * 1000)))

            global num_windows, total_time
            num_windows = num_windows + 1
            total_time = total_time + int(delta.total_seconds() * 1000)

            if num_windows == 1000:
                _LOGGER.info("Last 1000 data windows processed at an avg of {} msec".format(total_time//num_windows))
                num_windows = 0
                total_time = 0

            _LOGGER.info("-----------------------------------------------------------")

        data_window.append(elem)

    # _LOGGER.debug("vibration_features filter_ingest done")


from collections import defaultdict, Counter
from scipy.stats import kurtosis
from scipy.stats import skew
import scipy


def _kurtosis(x):
    return kurtosis(x)


def CPT5(x):
    den = len(x) * np.exp(np.std(x))
    return sum(np.exp(x)) / den


def SSC(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    xn_i1 = x[0:len(x) - 2]  # xn-1
    ans = np.heaviside((xn - xn_i1) * (xn - xn_i2), 0)
    return sum(ans[1:])


def wave_length(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    return sum(abs(xn_i2 - xn))


def norm_entropy(x):
    tresh = 2
    return sum(np.power(abs(x), tresh))


def SRAV(x):
    SRA = sum(np.sqrt(abs(x)))
    return np.power(SRA / len(x), 2)


def mean_abs(x):
    return sum(abs(x)) / len(x)


def calculate_entropy(list_values):
    counter_values = Counter(list_values).most_common()
    probabilities = [elem[1] / len(list_values) for elem in counter_values]
    entropy = scipy.stats.entropy(probabilities)
    return entropy


def calculate_statistics(list_values):
    n5 = 0.0 # np.nanpercentile(list_values, 5)
    n25 = 0.0 # np.nanpercentile(list_values, 25)
    n75 = 0.0 # np.nanpercentile(list_values, 75)
    n95 = 0.0 # np.nanpercentile(list_values, 95)
    median = np.nanpercentile(list_values, 50)
    mean = np.nanmean(list_values)
    std = np.nanstd(list_values)
    var = np.nanvar(list_values)
    rms = np.nanmean(np.sqrt(list_values ** 2))
    # New features
    kur = kurtosis(list_values)
    MeanAbs = 0.0 # mean_abs(list_values)
    norm_ent = 0.0 # norm_entropy(list_values)
    skewness = 0.0 # skew(list_values)
    CPT_5 = 0.0 # CPT5(list_values)
    SSC_1 = 0.0 # SSC(list_values)
    WL = 0.0 # wave_length(list_values)
    SRAV_1 = 0.0 # SRAV(list_values)
    return [n5, n25, n75, n95, median, mean, std, var, rms, kur, MeanAbs, norm_ent, skewness, CPT_5, SSC_1, WL, SRAV_1]


def calculate_crossings(list_values):
    zero_crossing_indices = np.nonzero(np.diff(np.array(list_values) > 0))[0]
    no_zero_crossings = len(zero_crossing_indices)
    mean_crossing_indices = np.nonzero(np.diff(np.array(list_values) > np.nanmean(list_values)))[0]
    no_mean_crossings = len(mean_crossing_indices)
    return [no_zero_crossings, no_mean_crossings]


# ftr_cols = ['median', 'mean', 'std_dev', 'variance', 'RMS', 'kurtosis', 'skewness', 'wavelength']

def get_features(list_values):
    statistics = calculate_statistics(list_values)
    features = statistics[4:10] + statistics[12:13]
    return features
